"""Default constants for NanoCodec modules."""

from jax import numpy as jnp

DEFAULT_LEAKY_RELU_NEGATIVE_SLOPE: float = 0.01
DEFAULT_AUDIO_DECODER_INPUT_CONV_SIZE: int = 7
DEFAULT_AUDIO_DECODER_OUTPUT_CONV_SIZE: int = 3
DEFAULT_AUDIO_DECODER_RESBLOCK_KERNEL_SIZES = (3, 7, 11)
DEFAULT_AUDIO_DECODER_RESBLOCK_DILATIONS = (1, 3, 5)
DEFAULT_FSQ_EPS: float = 1e-3

DEFAULT_NANOCODEC_PRECISION = jnp.float32
